package lsedit;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Date;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.Toolkit;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.FileOutputStream;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.PrintStream;
import java.io.BufferedReader;
import java.io.InputStreamReader;

import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JPanel;

class ClusterInterface extends JDialog implements ActionListener {

/*
    static String			m_default_command       = "c:/j2sdk1.4.2_04/bin/java.exe -classpath . lsedit.LsClusterer  /temp/ta /temp/ca";
	static String			m_default_export        = "/temp/ta";
	static String			m_default_import        = "/temp/ca";
*/
    static String			m_default_command       = "c:/j2sdk1.4.2_04/bin/java.exe -classpath . lsedit.LsClusterer";
	static String			m_default_export        = "";
	static String			m_default_import        = "";

	static boolean			m_default_delete_export = false;
	static boolean			m_default_delete_import = false;
	static boolean			m_default_leaves        = true;
	static boolean			m_default_feedback      = true;

	protected LandscapeEditorCore	m_ls;

	protected JTextField	m_command;
	protected JTextField	m_export;
	protected JTextField	m_import;		

	protected JCheckBox		m_delete_export;
	protected JCheckBox		m_delete_import;
	protected JCheckBox		m_leaves;
	protected JCheckBox		m_feedback;

	protected JLabel		m_message;

	protected JButton		m_ok;
	protected JButton		m_default;
	protected JButton		m_help;
	protected JButton		m_undo;
	protected JButton		m_cancel;

	// Working registers
	protected Diagram		m_dg;
	
	protected void log(String message)
	{
		if (m_feedback.isSelected()) {
			Date date = new Date();

			synchronized(this) {
				System.err.println(date.toLocaleString() + ": " + message);
	}	}	}

	protected void message(String string)
	{
		m_message.setText(string);
		log(string);
	}

	protected EntityInstance parentOfSet(Vector	set)
	{
		EntityInstance	e, parent, parent1;
		Enumeration		en;

		parent = null;
		for (en = set.elements(); en.hasMoreElements(); ) {
			e     = (EntityInstance) en.nextElement();
			parent1 = e.getContainedBy();
			if (parent == null) {
				parent = parent1;
			} else if (parent != parent1) {
				parent = null;
				break;
		}	}
		return(parent);
	}

	class ErrOutput implements Runnable {

		InputStream		m_inputStream = null;

		ErrOutput(InputStream inputStream)
		{
			m_inputStream = inputStream;
		}

		public void run()
		{

			BufferedReader		reader;
			InputStreamReader	isReader;
			String				s;


			BufferedInputStream bis;
			int c;

			try {
				isReader = new InputStreamReader(m_inputStream);
				reader   = new BufferedReader(isReader);

				while ((s = reader.readLine()) != null) {
					log(s);
				}
				reader.close();
			} catch (Exception error) {
				log("Syserr input error: " + error);
	}	}	}

	class WriteOutput implements Runnable {

		Process			m_process = null;
		InputStream		m_inputStream = null;
		boolean			m_waiting = true;

		Vector			m_selectedBoxes;

		WriteOutput(Vector selectedBoxes)
		{
			m_selectedBoxes = selectedBoxes;
		}

		public synchronized Process getProcess()
		{
			while (m_waiting) {
				try {
					wait(1);
				} catch(Exception error) {
					message("wait error: " + error);
				}	
			}
			return m_process;
		}

		public InputStream getInputStream()
		{
			return m_inputStream;
		}

		public Process startCommand(String command)
		{
			Process process = null;

			// Execute the command
			log("Executing [" + command + "]");

			try {
				Runtime runtime = Runtime.getRuntime();
				if (runtime == null) {
					message("No runtime available");
				} else {
					process = runtime.exec(command);
				}
			} catch (Exception error) {
				message("Exception executing [" + command + "] " + error);
				process = null;
			}	
			if (process != null) {
				ErrOutput	errOutput = new ErrOutput(process.getErrorStream());
				new Thread(errOutput).start();

				m_inputStream  = process.getInputStream();	// Stdout of process
			}
			m_process = process;
			m_waiting = false;
			return process;
		}

		public boolean write()
		{
			Vector			selectedBoxes = m_selectedBoxes;
			String			exportname    = m_export.getText();
			String			command       = m_command.getText();
			String			importname    = m_import.getText();
			Diagram			dg            = m_ls.getDiagram();


			Process			process    = null;
			File			exportfile = null;
			PrintStream		ps;

			try {
				OutputStream	os;

				if (exportname.length() == 0) {
					log("Piping output to subprocess");
					process = startCommand(command);
					if (process == null) {
						return false;
					}
					os = process.getOutputStream();
				} else {
					log("Exporting " + exportname);
		
					exportfile  = new File(exportname);
					os          = new FileOutputStream(exportfile);
				}
				BufferedOutputStream bos = new BufferedOutputStream(os);
				ps  = new PrintStream(bos);
			} catch (Exception error) {
				message("Exception creating output stream " + exportname + ": " + error);
				return(false);
			}

			Enumeration			en, en1;
			EntityInstance		e, e1;
			RelationInstance	ri;
			RelationClass		rc;
			int					cIndex;
			boolean				leaves = m_leaves.isSelected();

			try {
				ps.print("// Clusting TA file written by LSEdit " + Version.Number() + "\n\n");

				dg.writeSchemeTuples(ps);
				dg.writeSchemeAttributes(ps);
				ps.print("\n\nFACT TUPLE :\n\n");
				
				for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
					e = (EntityInstance) en.nextElement();
					e.writeInstance(ps);
					e.orMark(EntityInstance.SPRING_MARK);
				}

				for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
					e = (EntityInstance) en.nextElement();
					if (leaves) {
						// Use real edges to determine relationships
						for (en1 = e.srcRelationElements(); en1.hasMoreElements(); ) {
							ri = (RelationInstance) en1.nextElement();
							rc = ri.getRelationClass();
							// Consider only visible edges when drawing layout
							if (!rc.isClassVisible()) {
								continue;
							}
							e1 = ri.getDst();
							if (!e1.isMarked(EntityInstance.SPRING_MARK)) {
								continue;
							}
							ri.writeRelation(ps); 

						}
					} else {
						// Use lifted edges to determine relationships
						for (en = e.srcLiftedRelationElements(); en.hasMoreElements(); ) {
							ri = (RelationInstance) en.nextElement();
							rc = ri.getRelationClass();
							// Consider only visible edges when drawing layout
							if (!rc.isClassVisible()) {
								continue;
							}
							e1 = ri.getDrawDst();
							if (!e1.isMarked(EntityInstance.SPRING_MARK)) {
								continue;
							}
							ri.writeRelation(ps); 
				}	}	}

				ps.print("\n\nFACT ATTRIBUTE :\n\n"); 

				cIndex = dg.getCIndex();

				for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
					e = (EntityInstance) en.nextElement();
					e.writeAttribute(ps, cIndex);
					if (leaves) {
						// Use real edges to determine relationships
						for (en1 = e.srcRelationElements(); en1.hasMoreElements(); ) {
							ri = (RelationInstance) en1.nextElement();
							rc = ri.getRelationClass();
							// Consider only visible edges when drawing layout
							if (!rc.isClassVisible()) {
								continue;
							}
							e1 = ri.getDst();
							if (!e1.isMarked(EntityInstance.SPRING_MARK)) {
								continue;
							}
							ri.writeAttributes(ps); 

						}
					} else {
						// Use lifted edges to determine relationships
						for (en1 = e.srcLiftedRelationElements(); en1.hasMoreElements(); ) {
							ri = (RelationInstance) en1.nextElement();
							rc = ri.getRelationClass();
							// Consider only visible edges when drawing layout
							if (!rc.isClassVisible()) {
								continue;
							}
							e1 = ri.getDrawDst();
							if (!e1.isMarked(EntityInstance.SPRING_MARK)) {
								continue;
							}
							ri.writeAttributes(ps); 
				}	}	}

				ps.flush();
				ps.close();

				log("Export written");
			}
			catch(Exception error) {
				message("Exception writing output: " + error);
				return(false);
			}

			if (exportname.length() != 0 && command.length() != 0) {
				process = startCommand(command);
				if (process == null) {
					return false;
			}	}
			return true;
		}

		public void run()
		{
			write();
	}	}

	protected void transferNodes(EntityInstance parent, EntityInstance e /* other */)
	{
		Enumeration		en;
		EntityClass		ec;
		EntityInstance	e1    = null;
		EntityInstance	child = null;

		// This part builds the new containers
		// It is presumed that the input has no edges to/from containers

		for (en = e.getChildren(); en.hasMoreElements(); ) {
			if (child == null) {
				e1 = m_dg.updateClusterEntity(parent, e);
			}
			child = (EntityInstance) en.nextElement();
			transferNodes(e1, child);
		}
		if (child == null) {
			// This is a leaf node in the import
			String			id    = e.getId();
			EntityInstance	match = m_dg.getCache(id);
			e1 = m_dg.updateImportEntity(parent, e, match);
			e1.nandMark(EntityInstance.SPRING_MARK);
		}
		// Signal that parent being used
		parent.nandMark(EntityInstance.SPRING_MARK);
	}

	boolean waitFor(Process process) 
	{
		try {
			int	ret = process.waitFor();
			log("Process returned exit value of " + ret);
		} catch (Exception error) {
			message("WaitFor failed: " + error);
			return false;
		}
		return true;
	}

	protected boolean doit()
	{
		String	exportname = m_export.getText();
		String  command    = m_command.getText();
		String	importname = m_import.getText();

		if (exportname.length() == 0 && command.length() == 0) {
			message("Please specify an export file to write to or command to pipe to");
			return(false);
		}

		Diagram			dg = m_ls.getDiagram();
		EntityInstance	parent;

		m_dg = dg;

		// get user's selection of boxes to be laid out

		Vector masterBoxes = dg.getGroup();
		if (masterBoxes == null) {
		  message("No group selected");
		  return(false);
		}

		parent = parentOfSet(masterBoxes);
		if (parent == null) {
			message("Selected items do not have the same parent entity");
			return(false);
		}

		Vector				selectedBoxes;
		Enumeration			en;
		EntityInstance		e;
		boolean				leaves = m_leaves.isSelected();

		if (!leaves) {
			selectedBoxes = masterBoxes;
		} else {
			selectedBoxes = new Vector();
			for (en = masterBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.gatherLeaves(selectedBoxes);
		}	}

		WriteOutput	writeOutput = new WriteOutput(selectedBoxes);

		if (command.length() == 0) {
			writeOutput.write();
			message("Export written to '" + exportname + "'. No command specified");
			return(false);
		}

		Ta		ta     = new Ta(m_ls);
		
		InputStream	is;
		Process		process;
		String		result;
		int			ret;

		// Import the result

		if (importname.length() == 0) {
			new Thread(writeOutput).start();
			process = writeOutput.getProcess();
			if (process == null) {
				return false;
			}
			log("Reading subprocess output as our input");
			is = writeOutput.getInputStream();

			result = ta.loadTA("", is);
			if (result != null) {
				message(result);
				return(false);
			}
			log("Import loaded");
			if (!waitFor(process)) {
				return false;
			}

		} else {
			if (!writeOutput.write()) {
				return false;
			}
			process = writeOutput.getProcess();
			if (process == null) {
				return false;
			}
			if (!waitFor(process)) {
				return false;
			}
		
			File	importFile;
			log("Importing '" + importname);
			result = ta.loadTA(importname, null);
			if (result != null) {
				message(result);
				return(false);
			}
			log("Import loaded");

			if (m_delete_import.isSelected()) {
				try {
					File importfile = new File(importname);

					if (!importfile.delete()) {
						message("Unable to delete '" + importfile + "'");
					} else {
						message("Deleted " + importfile);
					}
				} 
				catch (Exception error) {
					message("Exception deleting '" + importname + "' " + error);
		}	}	}

		if (exportname.length() != 0 && m_delete_export.isSelected()) {
			try {
				File exportfile = new File(exportname);

				if (!exportfile.delete()) {
					message("Unable to delete '" + exportfile + "'");
				} else {
					log("Deleted " + exportfile);
			}	}
			catch (Exception error) {
				message("Exception deleting '" + exportname + "' " + error);
		}	}

		log("Reclustering");

		if (leaves) {
			for (en = masterBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.orMark(EntityInstance.SPRING_MARK);
		}	}

		dg.beginUndoRedo("Cluster Interface");

		EntityInstance root = ta.getRootInstance();
		EntityInstance child;

		for (en = root.getChildren(); en.hasMoreElements(); ) {
			child = (EntityInstance) en.nextElement();
			transferNodes(parent /* My TA */, child /* Other TA */);
		}

		// Make sure that we leave things as entered
		for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.nandMark(EntityInstance.SPRING_MARK);
		}

		if (leaves) {
			log("Removing redundant containers");

			/* These items were marked and are not things in selectedBoxes
			 * and have had nothing added back into/under them if SPRING_MARK
			 * still set
			 */

			for (en = masterBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				if (e.isMarked(EntityInstance.SPRING_MARK)) {
					dg.updateCutEntity(e);
		}	}	}

		dg.endUndoRedo();

		dg.redrawDiagram();

		message("Clustering operation complete");

		return(true);
	}

	protected String undoLabel()
	{
		LandscapeEditorCore	ls = m_ls;

		if (ls.getDiagram().undoEnabled()) {
			return "Disable undo";
		}
		return "Enable undo";
	}

	public ClusterInterface(LandscapeEditorCore ls)
	{
		super(ls.getFrame(), "Invoke external clustering tool", true);

		Container			contentPane;
		Font				font, bold;
		JLabel				label;
		JTextField			textfield;
		JCheckBox			checkbox;
		String				string;

		m_ls = ls;
		font = FontCache.getDialogFont();
		bold = font.deriveFont(Font.BOLD);

		setForeground(ColorCache.get(0,0,0));
		setBackground(ColorCache.get(192,192,192));
		setFont(font);

		JPanel topPanel    = new JPanel();
		JPanel labelPanel  = new JPanel();
		JPanel valuePanel  = new JPanel();
	

		GridLayout gridLayout;

		topPanel.setLayout(new BorderLayout());
		gridLayout = new GridLayout(3, 1, 0, 10);
		labelPanel.setLayout(gridLayout);
		gridLayout = new GridLayout(3, 1, 0, 10);
		valuePanel.setLayout(gridLayout);
	
		label = new JLabel("Command to be executed:", JLabel.RIGHT);
		label.setFont(bold);
		labelPanel.add(label);
		m_command = textfield = new JTextField(m_default_command, 80);
		textfield.setFont(font);
		valuePanel.add(textfield);

		label = new JLabel("Exported file used by command:", JLabel.RIGHT);
		label.setFont(bold);
		labelPanel.add(label);
		m_export = textfield = new JTextField(m_default_export, 80);
		textfield.setFont(font);
		valuePanel.add(textfield);

		label = new JLabel("Imported file created by command:", JLabel.RIGHT);
		label.setFont(bold);
		labelPanel.add(label);
		m_import = textfield = new JTextField(m_default_import, 80);
		textfield.setFont(font);
		valuePanel.add(textfield);

		JPanel optionsPanel = new JPanel();
		optionsPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

		m_delete_export = checkbox = new JCheckBox();
		checkbox.setFont(font);
		checkbox.setText("Delete export file");
		checkbox.setSelected(m_default_delete_export);
		optionsPanel.add(checkbox);
	
		m_delete_import = checkbox = new JCheckBox();
		checkbox.setFont(font);
		checkbox.setText("Delete import file");
		checkbox.setSelected(m_default_delete_import);
		optionsPanel.add(checkbox);

		m_leaves = checkbox = new JCheckBox();
		checkbox.setFont(font);
		checkbox.setText("Cluster leaves");
		checkbox.setSelected(m_default_leaves);
		optionsPanel.add(checkbox);
	
		m_feedback = checkbox = new JCheckBox();
		checkbox.setFont(font);
		checkbox.setText("Feedback");
		checkbox.setSelected(m_default_feedback);
		optionsPanel.add(checkbox);

		topPanel.add(BorderLayout.WEST, labelPanel);
		topPanel.add(BorderLayout.EAST, valuePanel);
		topPanel.add(BorderLayout.SOUTH, optionsPanel);

		contentPane = getContentPane();
		contentPane.add( BorderLayout.NORTH, topPanel );

		m_message = new JLabel(" ", JLabel.CENTER);

		if (ls.getDiagram().undoEnabled()) {
			string = "You might wish to disable undo/redo operations";
		} else {
			string = "You might wish to enable undo/redo operations";
		}

		m_message.setText(string);
		m_message.setFont(font);
		m_message.setForeground(Color.RED);
		m_message.setSize(400,50);
		m_message.setPreferredSize(new Dimension(400,50));
		contentPane.add( BorderLayout.CENTER, m_message);

		// --------------
		// Use a FlowLayout to center the button and give it margins.

		JPanel bottomPanel = new JPanel();

		bottomPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

		m_ok = new JButton("Ok");
		m_ok.setFont(bold);
		bottomPanel.add(m_ok);
		m_ok.addActionListener(this);

		m_default = new JButton("Default");
		m_default.setFont(bold);
		bottomPanel.add(m_default);
		m_default.addActionListener(this);

		m_undo = new JButton(undoLabel());
		m_undo.setFont(bold);
		bottomPanel.add(m_undo);
		m_undo.addActionListener(this);

		m_help = new JButton("Help");
		m_help.setFont(bold);
		bottomPanel.add(m_help);
		m_help.addActionListener(this);

		m_cancel = new JButton("Cancel");
		m_cancel.setFont(bold);
		bottomPanel.add(m_cancel);
		m_cancel.addActionListener(this);

		contentPane.add( BorderLayout.SOUTH, bottomPanel);

		// Resize the window to the preferred size of its components
		pack();
		show();
	}

	// ActionListener interface

	public void actionPerformed(ActionEvent ev)
	{
		Object	source;


		source = ev.getSource();

		if (source == m_default) {
			m_command.setText(m_default_command);
			m_export.setText(m_default_export);
			m_import.setText(m_default_import);
			m_delete_export.setSelected(m_default_delete_export);
			m_delete_import.setSelected(m_default_delete_import);
			m_feedback.setSelected(m_default_feedback);
			return;
		}

		if (source == m_help) {
			JOptionPane.showMessageDialog(m_ls.getFrame(), 	
			  "To merely export a file for later clustering provide an export filename.\n" +
			  "To run an external clusterer provide the command that invokes that external tool.\n" +
			  "If no export file specifying what is to be clustered is provided, this\n" +
			  "information will be piped directly to the specified commands standard input.\n" +
			  "The provided command is expected to write instructions as to how clustering is to\n" +
			  "be performed to the specified import file.  If no such file is specified it will\n" +
			  "be presumed that this input is to be read directly from the provided commands\n" +
			  "standard output stream.  Temporary files that are created may optionally be deleted\n" +
			  "when the clustering interface has no further use for them.\n" +
			  "\n" +
			  "The export file contains TA specifying the nodes to be clustered and the edges\n" +
			  "between them considered relevant when performing that clustering.  The import file\n" +
			  "should contain suitably structured TA showing how these exported nodes are to be\n" +
			  "clustered.  New containment hierarchies in the imported data will be constructed\n" +
			  "and the relevant nodes contained within such imported hierarchies moved into these\n" +
			  "hierarchies.  If sizing and positioning information is provided in the imported data\n" +
			  "this information will be mirrored in the resulting internal TA." 
				  , "Help", JOptionPane.OK_OPTION);
			return;
		}

		if (source == m_undo) {
			LandscapeEditorCore	ls = m_ls;
			ls.invertUndo();
			m_undo.setText(undoLabel());
			m_message.setText("");
			return;
		}

		if (source == m_ok) {
			boolean ret;

			m_ls.setCursor(Cursor.WAIT_CURSOR);
			ret = doit();
			m_ls.setCursor(Cursor.DEFAULT_CURSOR);

			if (!ret) {
				return;
			}
			m_default_command       = m_command.getText();
			m_default_export        = m_export.getText();
			m_default_import        = m_import.getText();
			m_default_delete_export = m_delete_export.isSelected();
			m_default_delete_import = m_delete_import.isSelected();
			m_default_feedback      = m_feedback.isSelected();
		} else if (source != m_cancel) {
			return;
		}
		setVisible(false);
		return;
	}

	public static void execute(LandscapeEditorCore ls)
	{
		Diagram	dg       = ls.getDiagram();
		Vector	selected = dg.getGroup();
		if (selected == null) {
			JOptionPane.showMessageDialog(ls.getFrame(), 	
			  "Select the items to be clustered"
				  , "No entities selected", JOptionPane.OK_OPTION);
		  return;
		}

		ClusterInterface dialog = new ClusterInterface(ls);
		dialog.dispose();
	}
} 





