package lsedit;

import java.util.Enumeration;
import java.util.Vector;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.Toolkit;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JPanel;

public class ClusterLayout extends LandscapeLayouter implements ToolBarEventHandler {

	static protected final int ATTRACTIVE_FORCE = 0;
	static protected final int SPARSE_FACTOR    = 1;
	static protected final int REPULSIVE_FORCE  = 2;
	static protected final int REPULSIVE_DIAMETER = 3;
	static protected final int ITERATIONS       = 4;
	static protected final int TIMEOUT          = 5;

	static protected final int FORM_CLUSTERS    = 6;
	static protected final int SEPARATION_FACTOR= 7;
	static protected final int MARGIN           = 8;
	static protected final int GAP              = 9;

	static protected final int REMOVE_CONTAINS  = 10;
	static protected final int MUSTBE_RELATED   = 11;
	static protected final int FEEDBACK         = 12;
	static protected final int LAST             = 13;

	static protected final double default_attractive_force = 0.05;
	static protected final double default_sparse_factor    = 1.0;
	static protected final double default_repulsive_force  = 0.01;
	static protected final double default_repulsive_diameter = 0.75;
	static protected final int    default_iterations       = 1000;
	static protected final int	  default_timeout           = 300;


	static protected final double default_separation_factor = 2.5;
	static protected final double  default_margin           = 0.05;
	static protected final double  default_gap              = 0.4;
	static protected final int     default_form_clusters    = 0; 
	static protected final boolean default_remove_contains  = true;
	static protected final boolean default_mustbe_related   = true;
	static protected final boolean default_feedback         = true;

	protected double	m_attractive_force = default_attractive_force;
	protected double	m_sparse_factor    = default_sparse_factor;
	protected double	m_repulsive_force  = default_repulsive_force;
	protected double	m_repulsive_diameter = default_repulsive_diameter;
	protected int		m_iterations       = default_iterations;
	protected int		m_timeout          = default_timeout;

	protected boolean	m_remove_contains  = default_remove_contains;
	protected boolean	m_mustbe_related   = default_mustbe_related;
	protected boolean	m_feedback         = default_feedback;

	protected double	m_separation_factor= default_separation_factor;
	protected double	m_margin           = default_margin;
	protected double	m_gap              = default_gap;
	protected int  	    m_form_clusters    = default_form_clusters;

	protected static String[] parameter_names = {
	 "Related force",
	 "Sparseness factor",
	 "Unrelated force",
	 "Repulsive range",
	 "Iterations",
	 "Timeout",
	 "Number clusters",
	 "Separation factor",
	 "Margin",
	 "Gap",
	 "Cluster leaves",
	 "Must be related",
	 "Provide feedback"

	};

	class ClusterConfigure extends JDialog implements ActionListener {

		protected JTextField[]		m_textfields;
		protected JCheckBox[]		m_checkboxes;

		protected JLabel			m_message;

		protected JButton			m_ok;
		protected JButton			m_default;
		protected JButton			m_cancel;
		protected JButton			m_help;
		protected boolean			m_ret;
		
		public ClusterConfigure(ClusterLayout layout)
		{
			super(layout.getLs().getFrame(), layout.getName() + " Configuration", true);

			Container			contentPane;
			Font				font, bold;
			JLabel				label;
			int					i, graph_size;
			String				string;
			boolean				bool = false;
			JTextField			textfield;
			JCheckBox			checkbox;

			m_ret        = false;
			graph_size   = TIMEOUT + 1;
			font         = FontCache.getDialogFont();
			bold         = font.deriveFont(Font.BOLD);

			setForeground(ColorCache.get(0,0,0));
			setBackground(ColorCache.get(192,192,192));
			setFont(font);

			JPanel topPanel    = new JPanel();
			JPanel labelPanel1 = new JPanel();
			JPanel valuePanel1 = new JPanel();
			JPanel labelPanel2 = new JPanel();
			JPanel valuePanel2 = new JPanel();
			JPanel labelPanel, valuePanel;

			GridLayout gridLayout;

			topPanel.setLayout(new BorderLayout());
			gridLayout = new GridLayout(8, 1, 0, 10);
			labelPanel1.setLayout(gridLayout);
			gridLayout = new GridLayout(8, 1, 0, 10);
			valuePanel1.setLayout(gridLayout);
			gridLayout = new GridLayout(8, 1, 0, 10);
			labelPanel2.setLayout(gridLayout);
			gridLayout = new GridLayout(8, 1, 0, 10);
			valuePanel2.setLayout(gridLayout);

			m_textfields = new JTextField[LAST];
			m_checkboxes = new JCheckBox[LAST];

			for (i = 0; i < LAST; ++i) {
				if (i < graph_size) {
					labelPanel = labelPanel1;
					valuePanel = valuePanel1;
				} else {
					labelPanel = labelPanel2;
					valuePanel = valuePanel2;
				}

				label = new JLabel(parameter_names[i], JLabel.RIGHT);
				label.setFont(bold);
				labelPanel.add(label);

				string = "";
				switch (i) {
				case ATTRACTIVE_FORCE:
					string += m_attractive_force;
					break;
				case SPARSE_FACTOR:
					string += m_sparse_factor;
					break;
				case REPULSIVE_FORCE:
					string += m_repulsive_force;
					break;
				case REPULSIVE_DIAMETER:
					string += m_repulsive_diameter;
					break;
				case ITERATIONS:
					string += m_iterations;
					break;
				case TIMEOUT:
					string += m_timeout;
					break;
				case SEPARATION_FACTOR:
					string += m_separation_factor;
					break;
				case MARGIN:
					string += m_margin;
					break;
				case GAP:
					string += m_gap;
					break;
				case REMOVE_CONTAINS:
					string = null;
					bool   = m_remove_contains;
					break;
				case FORM_CLUSTERS:
					string += m_form_clusters;
					break;
				case MUSTBE_RELATED:
					string = null;
					bool   = m_mustbe_related;
					break;
				case FEEDBACK:
					string = null;
					bool   = m_feedback;
				}

				if (string != null) {
					m_textfields[i] = textfield = new JTextField(string, 4);
					textfield.setFont(font);
					textfield.addActionListener(this);
					valuePanel.add(textfield);
				} else {
					m_checkboxes[i] = checkbox  = new JCheckBox();
					checkbox.setFont(font);
					checkbox.setSelected(bool);
					valuePanel.add(checkbox);
			}	}

			JPanel leftPanel   = new JPanel();
			leftPanel.setLayout(new BorderLayout());
			leftPanel.add(BorderLayout.WEST, labelPanel1);
			leftPanel.add(BorderLayout.EAST, valuePanel1);

			JPanel rightPanel = new JPanel();
			rightPanel.setLayout(new BorderLayout());
			rightPanel.add(BorderLayout.WEST, labelPanel2);
			rightPanel.add(BorderLayout.EAST, valuePanel2);
	
			topPanel.add(BorderLayout.WEST, leftPanel);
			topPanel.add(BorderLayout.EAST, rightPanel);

			contentPane = getContentPane();
			contentPane.add( BorderLayout.NORTH, topPanel );

			if (m_ls.getDiagram().undoEnabled()) {
				string = "You might wish to disable undo/redo operations";
			} else {
				string = "You might wish to enable undo/redo operations";
			}
			m_message = new JLabel(string, JLabel.CENTER);
			m_message.setFont(font);
			m_message.setForeground(Color.RED);
			m_message.setSize(400,50);
			m_message.setPreferredSize(new Dimension(400,50));
			contentPane.add( BorderLayout.CENTER, m_message);

			// --------------
			// Use a FlowLayout to center the button and give it margins.

			JPanel bottomPanel = new JPanel();

			bottomPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

			m_ok = new JButton("Ok");
			m_ok.setFont(bold);
			bottomPanel.add(m_ok);
			m_ok.addActionListener(this);

			m_default = new JButton("Default");
			m_default.setFont(bold);
			bottomPanel.add(m_default);
			m_default.addActionListener(this);

			m_cancel = new JButton("Cancel");
			m_cancel.setFont(bold);
			bottomPanel.add(m_cancel);
			m_cancel.addActionListener(this);


			m_help = new JButton("Help");
			m_help.setFont(bold);
			bottomPanel.add(m_help);
			m_help.addActionListener(this);

			contentPane.add( BorderLayout.SOUTH, bottomPanel);

			// Resize the window to the preferred size of its components
			pack();
			show();
		}

		public boolean ok()
		{
			return m_ret;
		}

		// ActionListener interface

		public void actionPerformed(ActionEvent ev)
		{
			Object			source;
			JTextField		textfield;
			JCheckBox		checkbox;
			double			dval;
			String			string;
			boolean			bool    = false;
			double[]	    values  = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
			int[]		    ivalues = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
			boolean[]       bools   = {false, false, false, false, false, false, false, false,
									   false, false, false, false, false, false, false, false,
									   false, false};
			int				ivalue; 

			source = ev.getSource();

			if (source == m_help) {
				JOptionPane.showMessageDialog(m_ls.getFrame(), 	
				  "Attractive force\n" +
				  "  Scales size of jumps towards ideal distance for related nodes.\n" +
				  "  Actual jump is proportional to difference from ideal distance\n" +
				  "Repulsive force\n" +
				  "  Scales size of jumps towards repulsive diameter for unrelated nodes\n" +
				  "  Actual jump is proportional to difference from repulsive diameter\n" +
				  "Sparse factor\n" +
				  "  The maximum width/height of any two distinct nodes is multiplied\n" +
				  "  by the sparse factor to give the ideal distance\n" +
				  "Repulsive diameter\n" +
				  "  Ideal ratio of distance between unrelated nodes to screen size\n" +
				  "Iterations\n" +
				  "  Number of times to iterate over the algorithm. Timeout overrides.\n" +
				  "Timeout\n" +
				  "  Maximum number of seconds to spend iterating\n"  +
				  "\n" +
				  "Number clusters\n" +
				  " Number of clusters to form. Use 0 to have the layouter use the\n" +
				  " separation factor to decide when to terminate clustering.\n"  +
				  " Set to 1 to view the graph used to cluster\n" +
				  "Separation factor\n" +
				  " Clustering stops when the minimum distance between nodes left\n" +
				  " to be clustered, is more than the prior minimum distance * factor.\n" +
				  "Margin\n" +
				  " Percentage of space to form margin within new clusters\n" +
				  "Gap\n" +
				  " Percentage of remaining space to act as separation between\n" +
				  " items in a new cluster\n" +
				  "Cluster leaves\n" +
				  " Cluster the leaves of the tree discarding all existing containers\n" +
				  "Mustbe related\n" +
				  " If true items added to a cluster must have an edge to something in\n" +
				  " that cluster.  Otherwise clustering is purely spacial"
				 	  , "Help", JOptionPane.OK_OPTION);
				return;
			}

			int		i;

			if (source == m_default) {
				for (i = 0; i < LAST; ++i) {
					string = "";
					switch (i) {
					case ATTRACTIVE_FORCE:
						string += default_attractive_force;
						break;
					case SPARSE_FACTOR:
						string += default_sparse_factor;
						break;
					case REPULSIVE_FORCE:
						string += default_repulsive_force;
						break;
					case REPULSIVE_DIAMETER:
						string += default_repulsive_diameter;
						break;
					case ITERATIONS:
						string += default_iterations;
						break;
					case TIMEOUT:
						string += default_timeout;
						break;
					case SEPARATION_FACTOR:
						string += default_separation_factor;
						break;
					case MARGIN:
						string += default_margin;
						break;
					case GAP:
						string += default_gap;
						break;
					case REMOVE_CONTAINS:
						string = null;
						bool   = default_remove_contains;
						break;
					case FORM_CLUSTERS:
						string += default_form_clusters;
						break;
					case MUSTBE_RELATED:
						string = null;
						bool   = default_mustbe_related;
						break;
					case FEEDBACK:
						string = null;
						bool   = default_feedback;
						break;
					}
					if (string != null) {
						m_textfields[i].setText(string);
					} else {
						m_checkboxes[i].setSelected(bool);
					}
				}
				return;
			}

			for (i = 0; i < LAST; ++i) {
				textfield = m_textfields[i];
				if (source == textfield || source == m_ok) {
					if (textfield == null) {
						checkbox = m_checkboxes[i];
						bools[i] = checkbox.isSelected();
					} else {
						string = textfield.getText();
						string.trim();
						switch (i) {
						case ITERATIONS:
						case TIMEOUT:
							try {
								ivalue = Integer.parseInt(string);
								if (ivalue <= 0) {
									m_message.setText(parameter_names[i] + " must be +ve");
									return;
								}
								ivalues[i] = ivalue;
							} catch (Throwable exception) {
								m_message.setText(parameter_names[i] + " not an integer value");
								return;
							}
							break;

						case FORM_CLUSTERS:
							try {
								ivalue = Integer.parseInt(string);
								if (ivalue < 0) {
									m_message.setText(parameter_names[i] + " may not be negative");
									return;
								}
								ivalues[i] = ivalue;
							} catch (Throwable exception) {
								m_message.setText(parameter_names[i] + " not an integer value");
								return;
							}
							break;
						default:
							try {
								values[i] = dval = Double.parseDouble(string);
								switch (i) {
								case MARGIN:
								case GAP:
									if (dval < 0.0 || dval >= 1.0) {
										m_message.setText(parameter_names[i] + " must be in the range 0 to 1.0");
										return;
								}	}

							} catch (Throwable exception) {
								m_message.setText(parameter_names[i] + " not a double precision value");
								return;
			}	}	}	}	}

			if (source != m_cancel && source != m_ok) {
				return;
			}

			if (source == m_ok) {
				m_attractive_force   = values[ATTRACTIVE_FORCE];
				m_sparse_factor      = values[SPARSE_FACTOR];
				m_repulsive_force    = values[REPULSIVE_FORCE];
				m_repulsive_diameter = values[REPULSIVE_DIAMETER];
				m_iterations         = ivalues[ITERATIONS];
				m_timeout            = ivalues[TIMEOUT];
				m_form_clusters      = ivalues[FORM_CLUSTERS];

				m_separation_factor  = values[SEPARATION_FACTOR];
				m_margin             = values[MARGIN];
				m_gap                = values[GAP];

				m_remove_contains    = bools[REMOVE_CONTAINS];
				m_mustbe_related     = bools[MUSTBE_RELATED];
				m_feedback           = bools[FEEDBACK];
				m_ret                = true;
			}
			setVisible(false);
			return;
		}
	}

	class ClusterNode {
		public EntityInstance	m_e;
		double					m_x;
		double					m_y;
		int						m_cluster;
		ClusterNode				m_next;
	};

	protected void log(String message)
	{
		if (m_feedback) {
			System.out.println(toLocaleString() + ": " + message);
	}	}

	public ClusterLayout(LandscapeEditorCore ls) 
	{
		super(ls);
	}

/****************************************************************************/

	public String getName()
	{
		return "Cluster";
	}

	public String getMenuLabel() 
	{
		return "Cluster";
	} 

/***************************************************************************/

	public boolean isConfigurable()
	{
		return true;
	}

	public boolean isLayouter()
	{
		return false;
	}

	public boolean configure(LandscapeEditorCore ls)
	{
		boolean ok;

		ClusterConfigure configure = new ClusterConfigure(this);
		ok = configure.ok();
		configure.dispose();
		return ok;
	}

	// N.B. can't use as layout tool during initial load

	public void doLayout1(Vector masterBoxes, EntityInstance container) 
	{
		Vector				selectedBoxes;
		int					size, i, j, iteration;
		EntityInstance		e, e1;
		Enumeration			en;
		RelationInstance	ri;
		RelationClass		rc;
		double				width, height, width1, height1;
		double				x, y, x1, y1, xmax, ymax, x1max, y1max, xdiff, radius, radius1, ydiff, f, length, newlength;
		double				ideal_length;
		long				timeout;


		// Do we want to cluster the selected items or the leaves of the selected items

		if (!m_remove_contains) {
			selectedBoxes = masterBoxes;
		} else {
			selectedBoxes = new Vector();
			for (en = masterBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.gatherLeaves(selectedBoxes);
		}	}

		size = selectedBoxes.size();
		if (size < 3) {
			// Not worth attempting to cluster
			return;
		}

		log("Clustering " + size + " items");

		// Use a variant of the spring layout algorithm to rapidly pull related things together
		// and push unrelated things apart. This differs from the spring layout algorithm in
		// being unconcerned about client/supplier relationships.

		ClusterNode[]	clusterNodes = new ClusterNode[size];
		ClusterNode		clusterNode, clusterNode1;
		boolean[][]		related;

		related = new boolean[size][];

		for (i = 0; i < size; ++i) {
			clusterNodes[i]       = clusterNode = new ClusterNode();
			related[i]            = new boolean[size-i];
			clusterNode.m_e       = e = (EntityInstance) selectedBoxes.elementAt(i);
			clusterNode.m_x       = e.xRelLocal();
			clusterNode.m_y       = e.yRelLocal();
			clusterNode.m_cluster = i;
			clusterNode.m_next    = null;
			e.orMark(EntityInstance.SPRING_MARK);
		}

		// Compute related and ideal length (max of two times any two widths or two heights -- converts 1.4 on diagonal)

		ideal_length   = 0;


		if (m_remove_contains) {
			// Use real edges to determine relationships

			for (i = 0; i < size; ++i) {
				clusterNode = clusterNodes[i];
				e           = clusterNode.m_e;
				for (en = e.srcRelationElements(); en.hasMoreElements(); ) {
					ri = (RelationInstance) en.nextElement();
					rc = ri.getRelationClass();
					// Consider only visible edges when drawing layout
					if (rc.isClassVisible()) {
						e1       = ri.getDst();
						if (e1.isMarked(EntityInstance.SPRING_MARK)) {
							for (j = i+1; j < size; ++j) {
								if (e1 == clusterNodes[j].m_e) {
									related[i][j-i] = true;
									break;
				}	}	}	}	}
				for (en = e.dstRelationElements(); en.hasMoreElements(); ) {
					ri = (RelationInstance) en.nextElement();
					rc = ri.getRelationClass();
					// Consider only visible edges when drawing layout
					if (rc.isClassVisible()) {
						e1       = ri.getSrc();
						if (e1.isMarked(EntityInstance.SPRING_MARK)) {
							for (j = i+1; j < size; ++j) {
								if (e1 == clusterNodes[j].m_e) {
									related[i][j-i] = true;
									break;
			}	}	}	}	}	}
		} else {
			// Use lifted edges to determine relationships
			for (i = 0; i < size; ++i) {
				clusterNode = clusterNodes[i];
				e           = clusterNode.m_e;
				for (en = e.srcLiftedRelationElements(); en.hasMoreElements(); ) {
					ri = (RelationInstance) en.nextElement();
					rc = ri.getRelationClass();
					// Consider only visible edges when drawing layout
					if (rc.isClassVisible()) {
						e1       = ri.getDrawDst();
						if (e1.isMarked(EntityInstance.SPRING_MARK)) {
							for (j = i+1; j < size; ++j) {
								if (e1 == clusterNodes[j].m_e) {
									related[i][j-i] = true;
									break;
				}	}	}	}	}
				for (en = e.dstLiftedRelationElements(); en.hasMoreElements(); ) {
					ri = (RelationInstance) en.nextElement();
					rc = ri.getRelationClass();
					// Consider only visible edges when drawing layout
					if (rc.isClassVisible()) {
						e1       = ri.getDrawSrc();
						if (e1.isMarked(EntityInstance.SPRING_MARK)) {
							for (j = i+1; j < size; ++j) {
								if (e1 == clusterNodes[j].m_e) {
									related[i][j-i] = true;
									break;
			}	}	}	}	}	}
		}


		// Compute ideal distance as the maximum length of two related nodes
		
		for (i = 0; i < size; ++i) {
			clusterNode = clusterNodes[i];
			e           = clusterNode.m_e;
			width       = e.widthRelLocal();
			height      = e.heightRelLocal();
			for (j = i+1; j < size; ++j) {
				if (!related[i][j-i]) {
					continue;
				}
				clusterNode1 = clusterNodes[j];
				e1           = clusterNode1.m_e;
				width1       = e1.widthRelLocal();
				height1      = e1.heightRelLocal();
				xdiff        = width + width1;
				if (xdiff > ideal_length) {
					ideal_length = xdiff;
				}
				ydiff        = height + height1;
				if (ydiff > ideal_length) {
					ideal_length = ydiff;
		}	}	}

		ideal_length  *= m_sparse_factor;

		log("Iterating over these " + size + " items");

		timeout = System.currentTimeMillis() + (m_timeout * 1000);
		for (iteration = m_iterations; iteration > 0; --iteration) {

			for (i = 0; i < size; ++i) {
				clusterNode = clusterNodes[i];
				x           = clusterNode.m_x;
				y           = clusterNode.m_y;

				for (j = i+1; j < size; ++j) {
					clusterNode1 = clusterNodes[j];
					x1           = clusterNode1.m_x;
					y1           = clusterNode1.m_y;
					xdiff        = x1 - x;
					ydiff        = y1 - y;

					if (xdiff == 0 && ydiff == 0) {
						// Choose some direction and make length none zero.
						xdiff = (i%3)-1;
						if (xdiff == 0) {
							ydiff = (j%2)*2-1;
						} else {
							ydiff = (j%3)-1;
					}	} 
					length		 = Math.sqrt(xdiff*xdiff + ydiff*ydiff);

					if (related[i][j-i]) {
						f = (length - ideal_length)         * m_attractive_force;
					} else {
						f = (length - m_repulsive_diameter) * m_repulsive_force;
					}
					x  += (f * xdiff/length);		// (xdiff/length) -> cos angle
					x1 -= (f * xdiff/length);
					y  += (f * ydiff/length);		// (ydiff/length) -> sin angle
					y1 -= (f * ydiff/length);
	
					clusterNode1.m_x = x1;
					clusterNode1.m_y = y1;
				}
				clusterNode.m_x  = x;
				clusterNode.m_y  = y;
			}
			if (System.currentTimeMillis() > timeout) {
				log("Timeout after " + (m_iterations - iteration) + " iterations");
				break;
		}	}

		log("Build graph for " + size + " items");

		ClusterNode	tail;
		int			clusters;
		Diagram		diagram = m_ls.getDiagram();


		clusters = size;
		if (m_form_clusters != 1  && m_form_clusters < clusters) {
			// Now compute the distances between 

			Vector		distances = new Vector();
			Distance	distance;
			int			pairs;
			
			for (i = 0; i < size; ++i) {
				clusterNode = clusterNodes[i];
				x          = clusterNode.m_x;
				y          = clusterNode.m_y;
				for (j = i+1; j < size; ++j) {
					if (m_mustbe_related && !related[i][j-i]) {
						continue;
					}
					clusterNode1 = clusterNodes[j];
					xdiff  = clusterNode1.m_x - x;
					ydiff  = clusterNode1.m_y - y;

					length = Math.sqrt(xdiff*xdiff + ydiff*ydiff);
					distances.add(new Distance(length, i, j));
			}	}

/*
			for (j = 0; j < size; ++j) {
				clusterNode = clusterNodes[j];
				if (clusterNode.m_cluster == j) {
					System.out.print(j + ") ");
					for (tail = clusterNode; tail != null; tail = tail.m_next) {
						System.out.print(" " + tail.m_e + "(" + tail.m_x + "x" + tail.m_y + ")");
					}
					System.out.println("");
			}	}
 */

			pairs = distances.size();

			log("Sorting " + pairs + " of distances");

			SortVector.byDistance(distances);
			ideal_length = -1;
			for (i = 0; i < pairs; ++i) {
				distance = (Distance) distances.elementAt(i);
				length   = distance.m_length;
				if (length > ideal_length * m_separation_factor) {
					if (m_form_clusters == 0 && ideal_length >= 0) {
						break;
				}	}
				ideal_length = length;
				clusterNode  = clusterNodes[distance.m_i];
				clusterNode1 = clusterNodes[distance.m_j];

				if (clusterNode.m_cluster == clusterNode1.m_cluster) {
					// Both already in same cluster
					continue;
				}
				j            = clusterNode.m_cluster;					// Cluster to add to
				clusterNode  = clusterNodes[j];							// Head of this cluster
				clusterNode1 = clusterNodes[clusterNode1.m_cluster];	// Head of this cluster

				// Put everything in clusterNode1 into clusterNode

				for (tail = clusterNode1; tail.m_next != null; tail = tail.m_next) {
					tail.m_cluster = j;
				}
				tail.m_cluster     = j;
				tail.m_next        = clusterNode.m_next;
				clusterNode.m_next = clusterNode1;

/*
				for (j = 0; j < size; ++j) {
					clusterNode = clusterNodes[j];
					if (clusterNode.m_cluster == j) {
						System.out.print(j + ") ");
						for (tail = clusterNode; tail != null; tail = tail.m_next) {
							System.out.print(" " + tail.m_e + "(" + tail.m_cluster + ")");
						}
						System.out.println("");
				}	}
 */
				--clusters;
				if (clusters <= m_form_clusters) {
					break;
				}
				if (clusters < 3) {
					break;
		}	}	}

		double		xm, ym, xc, yc, fill;
		int			cnt, size1;
		ClusterNode	utilities;
		String		text;

		// Putting loose things into a new container greatly simplifies layout problems
		// Putting them one at a time into the draw root and placing sensibly is very expensive

		size1     = size - 1;
		utilities = null;

		if (clusters > m_form_clusters) {
			log("Identifying utilities");

			size1     = -1;
			clusters  = 0;
			for (i = 0; i < size; ++i) {
				clusterNode = clusterNodes[i];

				if (clusterNode.m_cluster != i) {	// Not head of a cluster
					continue;
				}
				if (clusterNode.m_next == null) {
					if (utilities != null) {
						// Add cluster node to utilities
						clusterNode.m_cluster = utilities.m_cluster;
						clusterNode.m_next    = utilities.m_next;
						utilities.m_next      = clusterNode;
						continue;
					}
					utilities = clusterNode;
				}
				++clusters;
				size1 = i;
		}	}	

		log("Reorganising " + size + " items into " + clusters + " selected clusters");

		clusters = 0;
		e1       = null;
		for (i = 0; i <= size1; ++i) {
			clusterNode = clusterNodes[i];

			if (clusterNode.m_cluster == i) {	// Head of a cluster
				++clusters;

				e1 = diagram.updateNewEntity(null, container, null, false);
				text = "Cluster" + clusters;
				if (clusterNode == utilities) {
					if (m_form_clusters == 1) {
						text += " (Layout)";
					} else {
						text += " (Utilities)";
				}	}
				e1.setLabel(text);
				
				cnt = 0;
				x   = xmax = clusterNode.m_x;
				y   = ymax = clusterNode.m_y;
				for (tail = clusterNode; tail != null; tail = tail.m_next) {
					if (tail.m_x < x) {
						x = tail.m_x;
					}
					if (tail.m_y < y) {
						y = tail.m_y;
					}
					if (tail.m_x > xmax) {
						xmax = tail.m_x;
					}
					if (tail.m_y > ymax) {
						ymax = tail.m_y;
					}
					++cnt;
				}
				e1.setDescription("Cluster of " + cnt + " items");

				cnt   = (int) (Math.ceil(Math.sqrt(cnt)));
				// 
				width = (1.0 - m_margin) * (1.0 - m_gap) / cnt;
				xdiff = xmax - x;
				if (xdiff == 0) {
					xm = 0;
				} else {

					/* Solve:
					 * (m_margin/2)             = m*x    + c
					 * 1 - (m_margin/2) - width = m*xmax + c
					 */
					xm = (1.0 - m_margin - width)/xdiff;
				}
				xc  = (m_margin*0.5)-xm*x;

				ydiff = ymax - y;
				if (ydiff == 0) {
					ym = 0;
				} else {
					ym = (1.0 - m_margin - width)/ydiff;
				}
				yc = (m_margin*0.5)-ym*y;

//				System.out.println("xm=" + xm + " ym=" + ym + " xc=" + xc + " yc=" + yc);
				for (; clusterNode != null; clusterNode = clusterNode.m_next) {
					e  = clusterNode.m_e;
					x = clusterNode.m_x * xm + xc;
					y = clusterNode.m_y * ym + yc;
//					System.out.println(e + "=" + clusterNode.m_x + "x" + clusterNode.m_y + " -> " + x + "x" + y);

					e.updateRelLocal(x, y, width, width);
					diagram.updateMoveEntityContainment(e1, e);

					e.orMark(EntityInstance.SPRING_MARK);	// Incase cleared
			}	}
		}
			
		// Have now moved any leafs out of these containers so can safely delete
			
		if (m_remove_contains) {
			log("Removing obsolete containers");
			for (en = masterBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				if (!e.isMarked(EntityInstance.SPRING_MARK)) {
//					System.out.println("Deleting " + e);
					diagram.updateCutEntity(e);
		}	}	}

		for (i = 0; i < size; ++i) {
			clusterNode        = clusterNodes[i];
			e                 = clusterNode.m_e;
			e.nandMark(EntityInstance.SPRING_MARK);
		}

		if (clusters == 1 && e1 != null) {
			diagram.navigateTo(e1);
		} else {
			diagram.redrawDiagram();
		}	

		log("Finished forming " + clusters + " clusters");
		
	} // doLayout1


	public String doLayout(Diagram dg) 
	{
		EntityInstance parent;

		// get user's selection of boxes to be laid out


		Vector selectedBoxes = dg.getGroup();
		if (selectedBoxes == null) {
			  Util.beep();
			  return "No group selected";
		}

		String msg = allInDiagram(selectedBoxes);
		if (msg != null) {
			return msg;
		}

		parent = parentOfSet(selectedBoxes);
		if (parent == null) {
			return	"Cluster layout requires that all things laid out share same parent";
		}
		m_ls.setCursor(Cursor.WAIT_CURSOR);
		dg.beginUndoRedo("Cluster layout");
		doLayout1(selectedBoxes, parent);
		dg.endUndoRedo();
		m_ls.setCursor(Cursor.DEFAULT_CURSOR);

		return "Graph redrawn using Cluster Layout";
	} 

	public void processKeyEvent(int key, int modifiers, Object object) 
	{
		Diagram	dg;
		String	rmsg;

/*
		if (!configure(m_ls)) {
			return;
		}
 */

		dg = m_ls.getDiagram();
		if (dg != null) {
			rmsg = doLayout(dg);
			m_ls.doFeedback(rmsg);
			dg.rescaleDiagram();
			m_ls.repaintDg();
	}	}
} 





