package lsedit;

import java.util.Enumeration;
import java.util.Vector;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.Toolkit;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JComponent;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JButton;
import javax.swing.JTextField;
import javax.swing.JPanel;
import javax.swing.JLabel;

public class ExpandLayout extends LandscapeLayouter implements ToolBarEventHandler {

	class ExpandConfigure extends JDialog implements ActionListener, ItemListener {

		class ExpandImage extends JComponent
		{
			public ExpandImage()
			{
				super();

				Dimension	dimension = new Dimension(400,240);
				int			i;

				setLayout(null);
				setPreferredSize(dimension);
				setMinimumSize(dimension);
				setMaximumSize(dimension);
				setSize(dimension);
				setVisible(true);
			}
			
			public void paintComponent(Graphics g)
			{
				int	dimension, w, h;
				int	indentx, indenty, indentw, indenth;

				w = getWidth();
				h = getHeight();

				indentx = (int) (((double) w) * m_new_xborder);
				indenty = (int) (((double) h) * m_new_yborder);
				indentw = (int) (((double) w) * (1.0 - 2.0*m_new_xborder));
				indenth = (int) (((double) h) * (1.0 - 2.0*m_new_yborder));

				g.setColor(Color.black);
				g.drawRect(0,0,w-1, h-1);
				g.setColor(Color.blue);
				if (m_horizontal.isSelected()) {
					g.drawLine(indentx,           indenty, indentx,           indenty + indenth);
					g.drawLine(indentx + indentw, indenty, indentx + indentw, indenty + indenth);
				}
				if (m_vertical.isSelected()) {
					g.drawLine(indentx,          indenty,           indentx + indentw, indenty);
					g.drawLine(indentx,          indenty + indenth, indentx + indentw, indenty + indenth);
		}	}	}

		protected JTextField		m_xborder_textField;
		protected JTextField		m_yborder_textField;				
		protected ExpandImage		m_expandImage;

		protected JCheckBox			m_horizontal;
		protected JCheckBox			m_vertical;
		protected JLabel			m_message;

		protected JButton			m_ok;
		protected JButton			m_default;
		protected JButton			m_cancel;
		
		protected double			m_new_xborder;
		protected double			m_new_yborder;
		protected boolean			m_ret;


		protected boolean newXBorder()
		{
			String	string;
			double	border;

			string = m_xborder_textField.getText();
			try {
				border = Double.parseDouble(string);
				if (border < 0 || border > 0.5) {
					m_message.setText("X Border must be a fraction between 0 and 0.5");
					return false;
				}
			} catch (Throwable exception) {
				m_message.setText("X Border not a double precision string value");
				return false;
			}
			m_new_xborder = border;
			m_expandImage.repaint();
			return true;
		}

		protected boolean newYBorder()
		{
			String	string;
			double	border;

			string = m_yborder_textField.getText();
			try {
				border = Double.parseDouble(string);
				if (border < 0 || border > 0.5) {
					m_message.setText("Y Border must be a fraction between 0 and 0.5");
					return false;
				}
			} catch (Throwable exception) {
				m_message.setText("Y Border not a double precision string value");
				return false;
			}
			m_new_yborder = border;
			m_expandImage.repaint();
			return true;
		}

		public ExpandConfigure(ExpandLayout layout)
		{
			super(layout.getLs().getFrame(), layout.getName() + " Configuration", true);

			Container			contentPane;
			Font				font, bold;
			JLabel				label;


			m_ret  = false;
			font   = FontCache.getDialogFont();
			bold   = font.deriveFont(Font.BOLD);

			setForeground(ColorCache.get(0,0,0));
			setBackground(ColorCache.get(192,192,192));
			setFont(font);

			m_new_xborder      = m_xborder;
			m_new_yborder      = m_yborder;

			JPanel topPanel    = new JPanel();
			JPanel labelPanel  = new JPanel();
			JPanel valuePanel  = new JPanel();

			GridLayout gridLayout;

			topPanel.setLayout( new BorderLayout() );
			gridLayout = new GridLayout(4, 1, 0, 10);
			labelPanel.setLayout(gridLayout);

			gridLayout = new GridLayout(4,1, 0, 10);
			valuePanel.setLayout(gridLayout);

			label = new JLabel("X Border fraction of total:", JLabel.RIGHT);
			label.setFont(bold);
			labelPanel.add(label);

			m_xborder_textField = new JTextField("" + m_new_xborder, 10);
			m_xborder_textField.setFont(font);
			m_xborder_textField.addActionListener(this);
			valuePanel.add(m_xborder_textField);

			label = new JLabel("Y Border fraction of total:", JLabel.RIGHT);
			label.setFont(bold);
			labelPanel.add(label);

			m_yborder_textField = new JTextField("" + m_new_yborder, 10);
			m_yborder_textField.setFont(font);
			m_yborder_textField.addActionListener(this);
			valuePanel.add(m_yborder_textField);

			label = new JLabel("Expand Horizontally:", JLabel.RIGHT);
			label.setFont(bold);
			labelPanel.add(label);
			m_horizontal = new JCheckBox();
			m_horizontal.setSelected(m_xexpand);
			valuePanel.add(m_horizontal);
			m_horizontal.addItemListener(this);

			label = new JLabel("Expand Vertically:", JLabel.RIGHT);
			label.setFont(bold);
			labelPanel.add(label);
			m_vertical = new JCheckBox();
			m_vertical.setSelected(m_yexpand);
			valuePanel.add(m_vertical);
			m_vertical.addItemListener(this);


			topPanel.add( BorderLayout.WEST,   labelPanel);
			topPanel.add( BorderLayout.EAST,   valuePanel);

			contentPane = getContentPane();
			contentPane.add( BorderLayout.NORTH, topPanel );

			JPanel centerPanel = new JPanel();
			centerPanel.setLayout( new BorderLayout() );
			m_expandImage = new ExpandImage();
			centerPanel.add( BorderLayout.NORTH, m_expandImage);
			m_expandImage.validate();

			m_message = new JLabel(" ", JLabel.CENTER);
			m_message.setFont(font);
			m_message.setForeground(Color.RED);
			m_message.setSize(400,50);
			m_message.setPreferredSize(new Dimension(400,50));
			centerPanel.add( BorderLayout.SOUTH, m_message);

			contentPane.add( BorderLayout.CENTER, centerPanel);

			// --------------
			// Use a FlowLayout to center the button and give it margins.

			JPanel bottomPanel = new JPanel();

			bottomPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

			m_ok = new JButton("Ok");
			m_ok.setFont(bold);
			bottomPanel.add(m_ok);
			m_ok.addActionListener(this);

			m_default = new JButton("Default");
			m_default.setFont(bold);
			bottomPanel.add(m_default);
			m_default.addActionListener(this);

			m_cancel = new JButton("Cancel");
			m_cancel.setFont(bold);
			bottomPanel.add(m_cancel);
			m_cancel.addActionListener(this);

			contentPane.add( BorderLayout.SOUTH, bottomPanel);

			// Resize the window to the preferred size of its components
			pack();
			show();
		}

		public boolean ok()
		{
			return m_ret;
		}

		// ActionListener interface

		public void actionPerformed(ActionEvent ev)
		{
			Object			source;

			source = ev.getSource();

			if (source == m_default) {
				m_new_xborder = 0.025;
				m_new_yborder = 0.025;
				m_xborder_textField.setText("" + m_new_xborder);
				m_yborder_textField.setText("" + m_new_yborder);
				m_horizontal.setSelected(true);
				m_vertical.setSelected(true);
				m_expandImage.repaint();
				return;
			}

			if (source == m_xborder_textField) {
				newXBorder();
				return;
			}

			if (source == m_yborder_textField) {
				newYBorder();
				return;
			}

			if (source != m_cancel && source != m_ok) {
				return;
			}

			if (source == m_ok) {
				if (!newXBorder()) {
					return;
				}
				if (!newYBorder()) {
					return;
				}
				if (!m_horizontal.isSelected() && !m_vertical.isSelected()) {
					m_message.setText("Must specify expansion method");
					return;
				}

				m_xborder = m_new_xborder;
				m_yborder = m_new_yborder;
				m_xexpand = m_horizontal.isSelected();
				m_yexpand = m_vertical.isSelected();
				m_ret     = true;

			}
			setVisible(false);
			return;
		}

		public void itemStateChanged(ItemEvent ev)
		{
			m_expandImage.repaint();
		}
	}

	double	m_xborder   = 0.025;
	double	m_yborder   = 0.025;
	boolean m_xexpand   = true;
	boolean	m_yexpand   = true;

	public ExpandLayout(LandscapeEditorCore ls) 
	{
		super(ls);
	}

	public String getName()
	{
		return "Expand";
	}

	public String getMenuLabel() 
	{
		return "Expand layout to fill area";
	}

	public boolean isConfigurable()
	{
		return true;
	}

	public boolean isLayouter()
	{
		return true;
	}

	public boolean configure(LandscapeEditorCore ls)
	{
		boolean	ok;

		ExpandConfigure configure = new ExpandConfigure(this);
		ok = configure.ok();
		configure.dispose();
		return ok;
	}

	public void doLayout1(Vector selectedBoxes, EntityInstance container) 
	{
		Enumeration		en;
		EntityInstance	ce;
		boolean			first;
		double			min, max, max1, lastwidth, lastheight;
		double			xshift, xscale, yshift, yscale;
		double			oldx, oldy, newx, newy, width, height;

		if (!m_xexpand && !m_yexpand) {
			return;
		}

		xshift = 0;
		xscale = 1.0;
		yshift = 0.0;
		yscale = 1.0;
		min    = 0;
		max    = 0;
		lastwidth = 0;
		lastheight = 0;

		if (m_xexpand) {
			first = true;
			for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
				ce = (EntityInstance) en.nextElement();
				if (first) {
					min       = ce.xRelLocal();
					max       = min;
					lastwidth = ce.widthRelLocal();
					first     = false;
				} else {
					if (ce.xRelLocal() < min) {
						min = ce.xRelLocal();
					}
					max1 = ce.xRelLocal();
					if (max1 > max) {
						max       = max1;
						lastwidth = ce.widthRelLocal();
			}	}	}

			if (!first) {
				/*
					Want min -> m_xborder
					Want max -> 1 - m_xborder
				*/

				xshift = min;
				max   -= min;
				if (max != 0) {
					xscale = (1.0 - 2.0*m_xborder - lastwidth)/max;
		}	}	}

		if (m_yexpand) {
			first = true;
			for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
				ce = (EntityInstance) en.nextElement();
				if (first) {
					min        = ce.yRelLocal();
					max        = min;
					lastheight = ce.heightRelLocal();
					first      = false;
				} else {
					if (ce.yRelLocal() < min) {
						min = ce.yRelLocal();
					}
					max1 = ce.yRelLocal();
					if (max1 > max) {
						max        = max1;
						lastheight = ce.heightRelLocal();
			}	}	}

			if (!first) {
				/*
					Want to centralise so want (min+shift) = 1 - (max+shift) 
				                               shift       = (1 - max - min)
				*/

				yshift = min;
				max   -= min;
				if (max != 0) {
					yscale = (1.0 - 2.0*m_yborder - lastheight) / max;
		}	}	}

		for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
			ce   = (EntityInstance) en.nextElement();
			oldx = ce.xRelLocal();
			oldy = ce.yRelLocal();
			newx = (oldx - xshift)*xscale + m_xborder;
			newy = (oldy - yshift)*yscale + m_yborder;
			if (newx != oldx || newy != oldy) {
				width  = ce.widthRelLocal();
				height = ce.heightRelLocal();
				ce.updateRelLocal(newx, newy, width, height);
		}	}
		return;
	}

  // The doLayout method executes the Coffman-Graham Layer Assignment
  // algorithm and the Sugiyama algorithm on the boxes selected.
  // Assumption: All boxes selected are in the same container.

	public String doLayout(Diagram dg) 
	{
		EntityInstance parent;

		// get user's selection of boxes to be laid out

		m_ls.setLayouter(this);
		m_ls.setCursor(Cursor.WAIT_CURSOR);

		Vector selectedBoxes = dg.getGroup();
		if (selectedBoxes == null) {
			  Util.beep();
			  return "No group selected";
		}

		String msg = allInDiagram(selectedBoxes);
		if (msg != null) {
			return msg;
		}

		parent = parentOfSet(selectedBoxes);
		if (parent == null) {
			return	"Expand algorithm requires that all things laid out share same parent";
		}
		dg.beginUndoRedo("Expand layout");
		doLayout1(selectedBoxes, parent);
		dg.endUndoRedo();

		m_ls.setCursor(Cursor.DEFAULT_CURSOR);

		return "Graph redrawn using Expand algorithm";
	} // doLayout

	public void processKeyEvent(int key, int modifiers, Object object) 
	{
		Diagram				dg;
		String				rmsg;

/*
		if (!configure(m_ls)) {
			return;
		}
*/

		dg = m_ls.getDiagram();
		if (dg != null) {
			rmsg = doLayout(dg);
			dg.redrawDiagram();
			m_ls.doFeedback(rmsg);
	}	}
}

